{-------------------------------------------------------------------}
{ DSBSTREE.INC                                                      }
{ Copyright (c) 1993, 1994  Julian M. Bucknall. All Rights Reserved }
{-------------------------------------------------------------------}

{-A FirstThat iterator for cloning a binary search tree}
function FailsTreeCloning(T : PAbstractContainer; aData : pointer;
                          NT : pointer) : boolean;
  var
    NewTree : PBinTree absolute NT;
    DummyCursor : TTreeCursor;
    NewData : pointer;
    NoDataError : boolean;
  begin
    FailsTreeCloning := true;
    NoDataError := true;
    with NewTree^ do
      begin
        if OwnsData then
             NoDataError := DupData(aData, NewData)
        else NewData := aData;
        if NoDataError then
          if Insert(DummyCursor, NewData) then
               FailsTreeCloning := false
          else if OwnsData then DisposeData(NewData);
      end;
  end;

{=TBinSearchTree======================================================
A binary search tree

A sorted binary tree where for any given data object, all data objects
in its left subtree are less than it, and all data objects in the
right subtree are greater than it. This ordering relies on the Compare
method to be overridden.

19Sep93 JMB
======================================================================}
function  TBinSearchTree.Clone(DataOwner : boolean;
                               NewCompare : TCompareFunc) : PAbstractContainer;
  var
    NewTree : PBinTree;
    SaveTravType : TTraversalType;
  begin
    NewTree := PBinTree(NewContainer(DataOwner));
    if Assigned(NewTree) then
      begin
        NewTree^.SetCompare(NewCompare);
        if (not IsEmpty) then
          begin
            SaveTravType := TravType;
            TravType := PostOrder;
            if (FirstThat(FailsTreeCloning, NewTree) <> 0) then
              begin
                TravType := SaveTravType;
                Dispose(NewTree, Done);
                NewTree := nil;
                Error(edsCloneError);
              end;
            TravType := SaveTravType;
          end;
      end;
    Clone := NewTree;
  end;
{--------}
function  TBinSearchTree.Delete (Cursor : TTreeCursor) : TTreeCursor;
  var
    Walker,
    LeftChild : TTreeCursor;
  begin
    if IsLeaf(Cursor) then
      begin
        Delete := 0;
        Error(edsDelInvalidHere);
      end
    else if IsLeaf(Left(Cursor)) or IsLeaf(Right(Cursor)) then
      Delete := inherited Delete(Cursor)
    else {both children exist}
      begin
        Walker := Right(Cursor);
        LeftChild := Left(Walker);
        while not IsLeaf(LeftChild) do
          begin
            Walker := LeftChild;
            LeftChild := Left(Walker);
          end;
        bsSwapData(Cursor, Walker);
        Delete := inherited Delete(Walker);
      end;
  end;
{--------}
function  TBinSearchTree.Insert(var Cursor : TTreeCursor; aData : pointer): boolean;
  begin
    if Search(Cursor, aData) then
      begin
        Error(edsInsertDup);
        Insert := false;
      end
    else
      Insert := inherited Insert(Cursor, aData);
  end;
{--------}
procedure TBinSearchTree.Join(Cursor : TTreeCursor; Tree : PBinTree);
  var
    Walker,
    Dummy  : TTreeCursor;
    Data   : pointer;
  begin
    if Assigned(Tree) then
      with Tree^ do
        begin
          Walker := Root;
          while not IsEmpty do
            begin
              Data := Examine(Walker);
              Walker := Delete(Walker);
              if not Self.Insert(Dummy, Data) then
                {nothing - it cannot happen};
            end;
          Dispose(Tree, Done);
        end;
  end;
{--------}
function  TBinSearchTree.NewContainer(DataOwner : boolean) : PAbstractContainer;
  var
    NC : PAbstractContainer;
  begin
    NC := New(PBinSearchTree, Init(DataOwner));
    if Assigned(NC) then
      begin
        NC^.SetCompare(Compare);
        NC^.SetDisposeData(DisposeData);
        NC^.SetDupData(DupData);
      end;
    NewContainer := NC;
  end;
{--------}
function TBinSearchTree.Replace(Cursor : TTreeCursor; aData : pointer) : pointer;
  begin
    Replace := Examine(Cursor);
    Delete(Cursor);
    Insert(Cursor, aData);
  end;
{--------}
function  TBinSearchTree.Search(var Cursor : TTreeCursor; aData : pointer) : boolean;
  var
    CompResult : integer;
    Data       : pointer;
    Walker     : TTreeCursor;
  begin
    Walker := Root;
    if IsLeaf(Walker) then
      Search := false
    else
      begin
        CompResult := Compare(aData, Examine(Walker));
        if      (CompResult < 0) then Walker := Left(Walker)
        else if (CompResult > 0) then Walker := Right(Walker);
        while (not IsLeaf(Walker)) and (CompResult <> 0) do
          begin
            CompResult := Compare(aData, Examine(Walker));
            if      (CompResult < 0) then Walker := Left(Walker)
            else if (CompResult > 0) then Walker := Right(Walker);
          end;
        Search := (CompResult = 0);
      end;
    Cursor := Walker;
  end;
{--------}
procedure TBinSearchTree.bsSwapData(OldCursor, NewCursor : TTreeCursor);
  var
    Data : pointer;
  begin
    Data := GetNode(OldCursor)^.Data;
    GetNode(OldCursor)^.Data := GetNode(NewCursor)^.Data;
    GetNode(NewCursor)^.Data := Data;
  end;
{---------------------------------------------------------------------}
