{-------------------------------------------------------------------}
{ DSLIST.INC                                                        }
{ Copyright (c) 1993, 1994  Julian M. Bucknall. All Rights Reserved }
{-------------------------------------------------------------------}

{-A FirstThat iterator for cloning a single linked list}
function FailsSListCloning(SL : PAbstractContainer; aData : pointer;
                           NSL : pointer) : boolean;
  var
    NewList : PLinkList absolute NSL;
    NewData : pointer;
    NoDataError : boolean;
  begin
    {Note: assumes that NewList^.IsAfterLast is true}
    FailsSListCloning := true;
    NoDataError := true;
    with NewList^ do
      begin
        if OwnsData then
             NoDataError := DupData(aData, NewData)
        else NewData := aData;
        if NoDataError then
          if InsertBefore(NewData) then
               FailsSListCloning := false
          else if OwnsData then DisposeData(NewData);
      end;
  end;

{-A FirstThat iterator for cloning a SORTED single linked list}
function FailsSListSortCloning(SL : PAbstractContainer; aData : pointer;
                               NSL : pointer) : boolean;
  var
    NewList : PLinkList absolute NSL;
    NewData : pointer;
    NoDataError : boolean;
  begin
    FailsSListSortCloning := true;
    NoDataError := true;
    with NewList^ do
      begin
        if OwnsData then
             NoDataError := DupData(aData, NewData)
        else NewData := aData;
        if NoDataError then
          if InsertSorted(NewData) then
               FailsSListSortCloning := false
          else if OwnsData then DisposeData(NewData);
      end;
  end;

{=TLinkList===========================================================
A simple linked list.

19Sep93 JMB
======================================================================}
constructor TLinkList.Init(DataOwner : boolean);
  begin
    if not inherited Init(8, DataOwner) then Fail;
    {These allocations always succeed}
    BF := acNewNode(nil);
    AL := acNewNode(nil);
    BF^.Link := AL;
    AL^.Link := nil;
    Cursor := BF;
    Count := 0;
    IsSorted := true;
  end;
{--------}
function  TLinkList.Clone(DataOwner : boolean;
                          NewCompare : TCompareFunc) : PAbstractContainer;
  var
    NewList : PLinkList;
    Failed  : boolean;
  begin
    NewList := PLinkList(NewContainer(DataOwner));
    if not Assigned(NewList) then
      Error(ConstructorError)
    else
      NewList^.SetCompare(NewCompare);
      if (not IsEmpty) then
        begin
          NewList^.SetAfterLast;
          if IsSorted then
               Failed := FirstThat(FailsSListSortCloning, NewList) <> nil
          else Failed := FirstThat(FailsSListCloning, NewList) <> nil;
          if Failed then
            begin
              Dispose(NewList, Done);
              NewList := nil;
              Error(edsCloneError);
            end;
        end;
    Clone := NewList;
  end;
{--------}
procedure TLinkList.Delete;
  var
    Temp : PNode;
  begin
    {$IFDEF DEBUG}
    Assert((not IsBeforeFirst) and (not IsAfterLast), 'TLinkList.Delete');
    {$ENDIF}
    Temp := Cursor^.Link;
    acDisposeNode(Cursor);
    Cursor := BF^.Link;
    BF^.Link := Cursor^.Link;
    Cursor^.Link := Temp;
    if IsEmpty then
      IsSorted := true;
  end;
{--------}
procedure TLinkList.Empty;
  begin
    if IsBeforeFirst then
      Next;
    while not IsAfterLast do
      Erase;
    while not IsEmpty do
      begin
        Prev;
        Erase;
      end;
    if InDone then
      begin
        acDisposeNode(BF);
        acDisposeNode(AL);
      end
    else
      begin
        BF^.Link := AL;
        AL^.Link := nil;
        Cursor := BF;
      end;
  end;
{--------}
procedure TLinkList.Erase;
  begin
    if OwnsData then
      DisposeData(Examine);
    Delete;
  end;
{--------}
function  TLinkList.Examine : pointer;
  begin
    {$IFDEF DEBUG}
    Assert((not IsBeforeFirst) and (not IsAfterLast), 'TLinkList.Examine');
    {$ENDIF}
    Examine := Cursor^.Data;
  end;
{--------}
function  TLinkList.FirstThat(Action : TIteratorFunc; ExtraData : pointer) : pointer;
  var
    FoundIt : boolean;
  begin
    FirstThat := nil;
    FoundIt := false;
    SetBeforeFirst;
    Next;
    while (not IsAfterLast) and (not FoundIt) do
      if Action(@Self, Examine, ExtraData) then
        begin
          FoundIt := true;
          FirstThat := Examine;
        end
      else
        Next;
  end;
{--------}
procedure TLinkList.ForEach(Action : TIteratorProc; ExtraData : pointer);
  begin
    SetBeforeFirst;
    Next;
    while not IsAfterLast do
      begin
        Action(@Self, Examine, ExtraData);
        Next;
      end;
  end;
{--------}
function TLinkList.InsertAfter(aData : pointer) : boolean;
  var
    Node : PNode;
  begin
    InsertAfter := false;
    {$IFDEF DEBUG}
    Assert(not IsAfterLast, 'TLinkList.InsertAfter');
    {$ENDIF}
    Node := acNewNode(aData);
    if Assigned(Node) then
      begin
        Node^.Link := BF^.Link;
        BF^.Link := Node;
        IsSorted := false;
        InsertAfter := true;
      end;
  end;
{--------}
function TLinkList.InsertBefore(aData : pointer) : boolean;
  var
    Node : PNode;
  begin
    InsertBefore := false;
    {$IFDEF DEBUG}
    Assert(not IsBeforeFirst, 'TLinkList.InsertBefore');
    {$ENDIF}
    Node := acNewNode(aData);
    if Assigned(Node) then
      begin
        Node^.Link := Cursor^.Link;
        Cursor^.Link := Node;
        IsSorted := false;
        InsertBefore := true;
      end;
  end;
{--------}
function TLinkList.InsertSorted(aData : pointer) : boolean;
  begin
    if not IsSorted then
      if IsBeforeFirst then
           InsertSorted := InsertAfter(aData)
      else InsertSorted := InsertBefore(aData)
    else {the list is sorted}
      begin
        if Search(aData) then
          begin
            Error(edsInsertDup);
            InsertSorted := false;
          end
        else
          begin
            InsertSorted := InsertBefore(aData);
            IsSorted := true;
          end;
      end;
  end;
{--------}
function  TLinkList.IsAfterLast : boolean;
  begin
    IsAfterLast := (Cursor = AL);
  end;
{--------}
function  TLinkList.IsBeforeFirst : boolean;
  begin
    IsBeforeFirst := (Cursor = BF);
  end;
{--------}
procedure TLinkList.Join(List : PLinkList);
  var
    JoinNode : PNode;
  begin
    {$IFDEF DEBUG}
    Assert(not IsAfterLast, 'TLinkList.Join');
    {$ENDIF}
    if Assigned(List) then
      begin
        if not List^.IsEmpty then
          begin
            with List^ do
              begin
                SetBeforeFirst;
                Next;
              end;
            JoinNode := List^.Cursor;
            with List^ do
              begin
                SetAfterLast;
                Prev;
              end;
            JoinNode^.Link := Cursor;
            Cursor := List^.Cursor;
            inc(Count, List^.Count);
            IsSorted := false;
            {patch up List to be empty}
            with List^ do
              begin
                Cursor := BF;
                Count := 0;
              end;
          end;
        Dispose(List, Done);
      end;
  end;
{--------}
function  TLinkList.NewContainer(DataOwner : boolean) : PAbstractContainer;
  var
    NC : PAbstractContainer;
  begin
    NC := New(PLinkList, Init(DataOwner));
    if Assigned(NC) then
      begin
        NC^.SetCompare(Compare);
        NC^.SetDisposeData(DisposeData);
        NC^.SetDupData(DupData);
      end;
    NewContainer := NC;
  end;
{--------}
procedure TLinkList.Next;
  var
    Temp : PNode;
  begin
    {$IFDEF DEBUG}
    Assert(not IsAfterLast, 'TLinkList.Next');
    {$ENDIF}
    Temp := Cursor;
    Cursor := BF^.Link;
    BF^.Link := Cursor^.Link;
    Cursor^.Link := Temp;
  end;
{--------}
procedure TLinkList.Prev;
  var
    Temp : PNode;
  begin
    {$IFDEF DEBUG}
    Assert(not IsBeforeFirst, 'TLinkList.Prev');
    {$ENDIF}
    Temp := Cursor^.Link;
    Cursor^.Link := BF^.Link;
    BF^.Link := Cursor;
    Cursor := Temp;
  end;
{--------}
function  TLinkList.Replace(aData : pointer) : pointer;
  begin
    {$IFDEF DEBUG}
    Assert((not IsBeforeFirst) and (not IsAfterLast), 'TLinkList.Replace');
    {$ENDIF}
    if IsSorted then
      begin
        Replace := Examine;
        Delete;
        if not InsertSorted(aData) then
          {do nothing - it cannot happen as the node exists};
      end
    else
      with Cursor^ do
        begin
          Replace := Data;
          Data := aData;
        end;
  end;
{--------}
function TLinkList.Search(aData : pointer) : boolean;
  var
    CompResult   : integer;
    MoveForwards : boolean;
    StillLooking : boolean;
    Found        : boolean;
  begin
    if IsSorted then
      begin
        if IsBeforeFirst then
          MoveForwards := true
        else if IsAfterLast then
          MoveForwards := false
        else
          begin
            CompResult := Compare(aData, Examine);
            if (CompResult = 0) then
              begin
                Search := true;
                Exit;
              end;
            MoveForwards := (CompResult > 0);
          end;
        if MoveForwards then
          begin
            CompResult := 1;
            while (CompResult > 0) do
              begin
                Next;
                if IsAfterLast then
                     CompResult := -1
                else CompResult := Compare(aData, Examine);
              end;
            Search := CompResult = 0;
          end
        else {move backwards}
          begin
            CompResult := -1;
            while (CompResult < 0) do
              begin
                Prev;
                if IsBeforeFirst then
                     CompResult := +1
                else CompResult := Compare(aData, Examine);
              end;
            if (CompResult = 0) then
              Search := true
            else
              begin
                Next;
                Search := false;
              end;
          end;
      end
    else {the list is not currently sorted, search from the start}
      begin
        SetBeforeFirst;
        StillLooking := true;
        Found := false;
        while StillLooking and (not Found) do
          begin
            Next;
            if IsAfterLast then
                 StillLooking := false
            else Found := (Compare(aData, Examine) = 0);
          end;
        Search := Found;
      end;
  end;
{--------}
procedure TLinkList.SetAfterLast;
  var
    NextLink,
    Temp : PNode;
  begin
    {for speed reasons, code from first principles,
     this is equivalent to:
       while not IsAfterLast do Next;}
    NextLink := BF^.Link;
    while (Cursor <> AL) do
      begin
        Temp := Cursor;
        Cursor := NextLink;
        NextLink := Cursor^.Link;
        Cursor^.Link := Temp;
      end;
    BF^.Link := NextLink;
  end;
{--------}
procedure TLinkList.SetBeforeFirst;
  var
    NextLink,
    Temp : PNode;
  begin
    {for speed reasons, code from first principles,
     this is equivalent to:
       while not IsBeforeFirst do Prev;}
    NextLink := BF^.Link;
    while (Cursor <> BF) do
      begin
        Temp := Cursor^.Link;
        Cursor^.Link := NextLink;
        NextLink := Cursor;
        Cursor := Temp;
      end;
    BF^.Link := NextLink;
  end;
{--------}
function  TLinkList.Split : PLinkList;
  var
    {$IFDEF MaxIs64KItems}
    TempCount : word;
    {$ELSE}
    TempCount : longint;
    {$ENDIF}
    NewList   : PLinkList;
    LastNodeLeftBehind,
    JoinNode  : PNode;
  begin
    {$IFDEF DEBUG}
    Assert((not IsBeforeFirst) and (not IsAfterLast), 'TLinkList.Split');
    {$ENDIF}
    NewList := PLinkList(NewContainer(OwnsData));
    if not Assigned(NewList) then
      Error(ConstructorError)
    else
      begin
        LastNodeLeftBehind := Cursor^.Link;

        TempCount := 0;
        JoinNode := Cursor;
        while not IsAfterLast do
          begin
            inc(TempCount);
            Next;
          end;

        JoinNode^.Link := NewList^.BF;
        NewList^.Cursor := AL^.Link;
        NewList^.Next;
        NewList^.Count := TempCount;
        NewList^.IsSorted := IsSorted;

        dec(Count, TempCount);
        AL^.Link := LastNodeLeftBehind;
        if IsEmpty then
          IsSorted := true;
      end;
    Split := NewList;
  end;
{---------------------------------------------------------------------}
